#include "pindmd2.h"
#include "dmd.h"

#include "lusb0_usb.h"
#include <stdio.h>

#define REPORT_MESSAGE printf

//define vendor id and product id
#define VID 0x0314
#define PID 0xe457

//configure and interface number
//TODO figure out what these mean
#define MY_CONFIG 1
#define MY_INTF 0

//endpoints for communication
#define EP_IN 0x81
#define EP_OUT 0x01

//maximum packet size, 64 for full-speed, 512 for high-speed
#define BUF_SIZE 2052//64

namespace
{
   const unsigned char DOT_COLOUR_LOOKUP_TABLE[NUM_DMD_COLOURS] =
   {
      0,
      1,
      1,
      1,
      2,
      2,
      2,
      4,
      4,
      4,
      8,
      9,
      10,
      11,
      12,
      15
   };

   char writeBuffer[ BUF_SIZE ];
   char readBuffer[1];
   //context handle for transfer and receive
   //think of them bit like session id:s
   void* asyncReadContext = NULL;
   void* asyncWriteContext = NULL;

   usb_dev_handle *device = NULL;

   //gets device handle
   //@param none
   //@return device handle
   usb_dev_handle* open_dev( void )
   {
      //contains usb busses present on computer
      struct usb_bus *bus;
      //contains devices present on computer
      struct usb_device *dev;
      //used to skip first device
      //bool first = true;
      //loop through busses and devices
      for (bus = usb_get_busses(); bus; bus = bus->next)
      {
         for (dev = bus->devices; dev; dev = dev->next)
         {
            //if device vendor id and product id are match
            if (dev->descriptor.idVendor == VID && dev->descriptor.idProduct == PID)
               return usb_open(dev);
         }
      }
      //return null if no devices were found
      return NULL;
   }

   //*****************************************************
   //* Name:			renderDMDFrame
   //* Purpose:
   //* In:
   //* Out:
   //*****************************************************
   void renderDMDFrame(int width, int height, const unsigned char *currbuffer)
   {
      int byteIdx=4;
      int bd0,bd1,bd2,bd3;
      int pixel;
      int i,j,v;
      int ret = 0;
      unsigned char frame_buf[2052];

      memset(frame_buf,0,2052);
      frame_buf[0] = 0x81; // frame sync bytes
      frame_buf[1] = 0xC3;
      frame_buf[2] = 0xE7;
      frame_buf[3] = 0x0;     // command byte (not used)

      // dmd height
      for (j = 0; j < height; j++)
      {
         // dmd width
         for (i = 0; i < width; i+=8)
         {
            bd0 = 0;
            bd1 = 0;
            bd2 = 0;
            bd3 = 0;
            for (v = 7; v >= 0; v--)
            {
               // pixel colour
               pixel = currbuffer[(j*128) + (i+v)];
               // 16 color mode hue remapping for proper gradient
               bd0 <<= 1;
               bd1 <<= 1;
               bd2 <<= 1;
               bd3 <<= 1;
               //bd4 <<= 1;
               if (pixel & 1)
                  bd0 |= 1;
               if (pixel & 2)
                  bd1 |= 1;
               if (pixel & 4)
                  bd2 |= 1;
               if (pixel & 8)
                  bd3 |= 1;
            }
            frame_buf[byteIdx]      = bd0;
            frame_buf[byteIdx+512]  = bd1;
            frame_buf[byteIdx+1024] = bd2;
            frame_buf[byteIdx+1536] = bd3;
            byteIdx++;
         }
      }

      //writeBuffer[0]=0;
      //usb_bulk_write(device, EP_OUT, writeBuffer, 1, 5000);
      //usb_bulk_read(device, EP_IN, readBuffer, 1, 5000);
      memcpy(writeBuffer,frame_buf,2052);
      usb_bulk_write(device, EP_OUT, writeBuffer, 2052, 5000);
      //printf("frame sent: %s\n",usb_strerror());


      /*memcpy(writeBuffer,frame_buf,2048);
       ret = usb_submit_async( asyncWriteContext, writeBuffer, sizeof( writeBuffer ) );
       if( ret < 0 )
        printf("error usb_submit_async:\n%s\n", usb_strerror());

       //wait for read and write transfers to complete or 5 seconds
       usb_reap_async( asyncWriteContext, 5000 );*/
   }
}

PINDMD2::PINDMD2(void)
:
device_enabled(false)
{
   //init usb library
   usb_init();
   //find busses
   usb_find_busses();
   //find devices
   usb_find_devices();

   //try to open our device
   device = open_dev();

   if (device)
   {
      REPORT_MESSAGE("Opened pindmd device\n");

      //set configuration
      if (usb_set_configuration( device, MY_CONFIG ) < 0)
      {
         //if not succesfull, notify
         REPORT_MESSAGE("Error setting pindmd configuration\n");
         //close device
         usb_close( device );
      }
      else
      {
         REPORT_MESSAGE("pindmd configuration set\n");

         //try to claim interface for use
         if (usb_claim_interface( device, MY_INTF ) < 0)
         {
            REPORT_MESSAGE("Error claiming pindmd interface\n");
            //close device and exit
            usb_close( device );
         }
         else
         {
            REPORT_MESSAGE("pindmd interface claimed\n");
            device_enabled = true;
         }
      }
   }
   else
   {
      REPORT_MESSAGE("Failed to open pindmd device\n");
   }
}

PINDMD2::~PINDMD2()
{
   if (device_enabled)
   {
      //clean up.
      //kitten dies every time you don't clean up after yourself.
      //free context handlers
      usb_free_async( &asyncWriteContext );
      //relese interface
      usb_release_interface( device, MY_INTF );
      //close device and exit
      usb_close( device );
   }
}

void PINDMD2::render_frame(const unsigned char* frame_data)
{
   if (device_enabled)
   {
      unsigned char buffer[DMD_WIDTH*DMD_HEIGHT];

      for (int dot_index = 0; dot_index < DMD_WIDTH*DMD_HEIGHT; dot_index++)
      {
         const unsigned char dot_src_value = frame_data[dot_index];
         const unsigned char dot_value = DOT_COLOUR_LOOKUP_TABLE[dot_src_value];
         buffer[dot_index] = dot_value;
      }

      renderDMDFrame(DMD_WIDTH, DMD_HEIGHT, buffer);
   }
}

bool PINDMD2::is_connected(void) const
{
   return device_enabled;
}

